# -*- coding: utf-8 -*-
"""
Created on Thu Nov  7 15:31:13 2024

@author: mstep
"""

#!/usr/bin/env python
# coding: utf-8

#%%
class PointData:
    def __init__(self, location, cluster=None):
        self.location = location
        self.cluster = cluster

class Cluster:
    def __init__(self, location, color):
        self.location = location
        self.color = color

# In[2]:

import tkinter as tk
import random
import math

class KMeansApp:
    def __init__(self):
        self.seeds = []
        self.points = []
        self.clusters = []

        # Cluster colors.
        self.colors = ['cyan', 'lime', 'yellow', 'magenta', 'orange',
                       'green', 'blue', 'red']
        self.num_colors = len(self.colors)

        self.window = tk.Tk()
        self.window.title('k_means')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry('700x500')

        # Left frame.
        frame = tk.Frame(self.window)
        frame.pack(side=tk.LEFT, anchor=tk.N, padx=3, pady=3)

        LABEL_WIDTH = 10
        ENTRY_WIDTH = 8
        BUTTON_WIDTH = 15

        # Seeds.
        num_seeds_entry, self.num_seeds_var = \
            self.make_field(frame, LABEL_WIDTH, "# Seeds:",
                            ENTRY_WIDTH, '3')

        # Points.
        _, self.num_points_var = \
            self.make_field(frame, LABEL_WIDTH, "# Points:",
                            ENTRY_WIDTH, '200')
        self.make_points_button = self.make_button(
            frame, BUTTON_WIDTH, 'Make Points', self.make_points_click,
            tk.NORMAL)

        # Delay (ms).
        label = tk.Label(frame, text='Delay (ms):')
        label.pack(side=tk.TOP, pady=(50,0))

        self.ms_scale = tk.Scale(frame, from_=0, to=1000,
            command=self.ms_changed, orient=tk.HORIZONTAL)
        self.ms_scale.pack(side=tk.TOP, fill=tk.X)
        self.ms = 500
        self.ms_scale.set(self.ms)

        # Clusters.
        _, self.num_clusters_var = \
            self.make_field(frame, LABEL_WIDTH, "# Clusters:",
                            ENTRY_WIDTH, '3')
        self.make_clusters_button = self.make_button(
            frame, BUTTON_WIDTH, 'Make Clusters', self.make_clusters_click,
            tk.NORMAL)

        # Output.
        self.result_label = tk.Label(frame, text='')
        self.result_label.pack(side=tk.TOP, padx=5, pady=5)

        # Canvas.
        self.canvas = tk.Canvas(self.window, bg='white', borderwidth=2,
                                relief=tk.SUNKEN)
        self.canvas.pack(side=tk.LEFT, fill=tk.BOTH,
                         expand=True, padx=(0,5), pady=5)

        num_seeds_entry.focus_force()
        self.window.mainloop()

    def make_field(self, parent, label_width, label_text,
                   entry_width, entry_text, entry_state=tk.NORMAL):
        '''
        Make a Label and associated Entry in columns 0 and 1.
        Return the Entry and its variable.
        '''
        frame = tk.Frame(parent)
        frame.pack(side=tk.TOP)

        label = tk.Label(frame, text=label_text, width=label_width,
            anchor=tk.W)
        label.pack(side=tk.LEFT, padx=3, pady=3)

        var = tk.StringVar(parent, value=entry_text)
        entry = tk.Entry(frame, textvariable=var, width=entry_width,
            justify=tk.RIGHT, state=entry_state)
        entry.pack(side=tk.LEFT, padx=3, pady=5)
        return entry, var

    def make_button(self, parent, button_width, text, command, state):
        '''Make a Button in this row spanning columns 0 and 1.'''
        button = tk.Button(parent, text=text, width=button_width,
                           command=command, state=state)
        button.pack(side=tk.TOP, padx=3, pady=3)
        return button

    def kill_callback(self):
        # Destroy the tkinter window.
        self.window.destroy()

    def make_points_click(self):
        # Make random points.
        num_seeds = int(self.num_seeds_var.get())
        num_points = int(self.num_points_var.get())
        self.seeds = []
        self.points = []
        self.clusters = []

        # Make the seeds.
        margin = 100
        xmin = margin
        xmax = self.canvas.winfo_width() - margin
        ymin = margin
        ymax = self.canvas.winfo_height() - margin
        for _ in range(num_seeds):
            x = random.randint(xmin, xmax)
            y = random.randint(ymin, ymax)
            self.seeds.append((x, y))

        # Make the points.
        margin = 50
        xmin = margin
        xmax = self.canvas.winfo_width() - margin
        ymin = margin
        ymax = self.canvas.winfo_height() - margin
        max_r = min(self.window.winfo_width(), self.window.winfo_height()) / 8
        for _ in range(num_points):
            # Pick a random seed.
            seed = random.choice(self.seeds)
            r = max_r * random.random() + max_r * random.random()
            angle = random.random() * (2 * math.pi)
            x = seed[0] + r * math.cos(angle)
            y = seed[1] + r * math.sin(angle)
            self.points.append(PointData((x, y)))

        # Draw.
        self.draw()

    def make_clusters_click(self):
        '''Find the clusters.'''
        self.result_label['text'] = ''
        self.num_iterations = 0

        # Pick random clusters.
        MARGIN = 50
        xmin = MARGIN
        xmax = self.canvas.winfo_width() - MARGIN
        ymin = MARGIN
        ymax = self.canvas.winfo_height() - MARGIN
        self.num_clusters = int(self.num_clusters_var.get())
        self.clusters = []
        for i in range(self.num_clusters):
            x = random.randrange(xmin, xmax)
            y = random.randrange(ymin, ymax)
            color = self.colors[i % self.num_colors]
            self.clusters.append(Cluster((x, y), color))

        # Display the first result.
        self.show_next_result()

    def ms_changed(self, _):
        self.ms = self.ms_scale.get()

    def show_next_result(self):
        '''Update the clusters and display the current result.'''
        self.num_iterations += 1
        self.result_label['text'] = f'Iteration: {self.num_iterations}'

        # Assign each point to a cluster.
        for point in self.points:
            # See which cluster is closest to this point.
            best_dist = math.inf
            point.cluster = None
            for cluster in self.clusters:
                test_dist = math.dist(point.location, cluster.location)
                if test_dist < best_dist:
                    best_dist = test_dist
                    point.cluster = cluster

        # Calculate the clusters' new centroids.
        for cluster in self.clusters:
            cluster.new_location = [0, 0]
            cluster.member_count = 0
        for point in self.points:
            point.cluster.new_location[0] += point.location[0]
            point.cluster.new_location[1] += point.location[1]
            point.cluster.member_count += 1
        for cluster in self.clusters:
            if cluster.member_count > 0:
                cluster.new_location[0] /= cluster.member_count
                cluster.new_location[1] /= cluster.member_count

        # See how far they moved.
        max_move = 0
        for cluster in self.clusters:
            # See how far it moved.
            move = math.dist(cluster.location, cluster.new_location)
            if move > max_move:
                max_move = move

            # Update it.
            cluster.location = cluster.new_location

        # See if the centroids moved substantially.
        if max_move > 1:
            # They moved at least a bit. Pause and repeat.
            self.window.after(self.ms, self.show_next_result)
        else:
            self.result_label['text'] = f'# Iterations: {self.num_iterations}'

        # Draw the solution.
        self.draw()

    def draw(self):
        '''Draw the points and clusters.'''
        # Delete any previous drawing objects.
        self.canvas.delete(tk.ALL)

        # Draw the points.
        radius = 5
        for point in self.points:
            if point.cluster is not None:
                color = point.cluster.color
            else:
                color = ''
            self.draw_point(point.location, radius, color)

        # Draw the seeds.
        radius = 10
        fill = 'hot pink'
        outline = 'black'
        for seed in self.seeds:
            points = [
                (seed[0] - radius, seed[1]),
                (seed[0], seed[1] - radius),
                (seed[0] + radius, seed[1]),
                (seed[0], seed[1] + radius),
            ]
            points.append(points[0])
            self.canvas.create_polygon(points, fill=fill)
            self.canvas.create_line(points, fill=outline, width=3)

        # Draw the clusters.
        radius = 10
        for cluster in self.clusters:
            self.draw_point(cluster.location, radius,
                            cluster.color, width=3)

    def draw_point(self, location, radius, fill, outline='black', width=1):
            x, y = location
            self.canvas.create_oval(
                x - radius, y - radius,
                x + radius, y + radius,
                fill=fill, outline=outline, width=width)

# In[3]:

KMeansApp()
print('Done')


# In[ ]:
